#include <iostream>
#include <string>
#include <vector>
#include <cassert>

using namespace std;

class AminoAcid {
    public:
        AminoAcid(); //Конструктор по умолчанию
        AminoAcid(const  vector <char> &newNuclio); //Конструктор инициализации
        AminoAcid(const AminoAcid &originAcid); //Конструктор копирования
        virtual void addNuclio(const char nuclio); //Функция добавления нуклеотида в кислоту
        void combineInAcid(); //Функция объединения в кислоту
        virtual void chowCode() const; //Функция демонстрации кода
        void chowAddedNuclio() const; //Функция демонстрации нуклео
        virtual string getCode() const; //Функция получения конечного кода кислоты
        vector <char> getNuclio() const; //Функция получения добавленных нуклеотидов
        virtual string identification() const;
    protected:
        virtual int tryAddNuclio(char nuclio); //Процесс добавления с проверками
        bool tryCombineInAcid(); //Процесс объединения в кислоту с проверками
        const vector <char> NUCLIOTIDES = {'A', 'C', 'G', 'U'}; //Возможные нуклеотиды
    private:
        string totalCode = ""; //Конечный код кислоты
        vector <char> addedNuclio; //Добавленные нуклеотиды
};

string AminoAcid::identification() const {
    return "AminoAcid";
}

AminoAcid::AminoAcid() {}

AminoAcid::AminoAcid(const vector <char> &newNuclio) {
    int numberNuclio = 0;
    while(addedNuclio.size() != 3 && numberNuclio != newNuclio.size()){
        tryAddNuclio(newNuclio[numberNuclio]);
        numberNuclio++;
    }
    if (addedNuclio.size() == 0){
        cout << "Кислота создана, но из заданных нуклеотидов не добавлен ни один!" << endl;
    } else{
        cout << "Кислота создана. Она состоит из " << addedNuclio.size() << " элементов: ";
        for (int i = 0; i < addedNuclio.size(); i++) {
            cout << " " << addedNuclio[i];
        }
        cout << endl;
    }
}

AminoAcid::AminoAcid(const AminoAcid &originAcid) {
    totalCode = originAcid.getCode();
    for (int i = 0; i < originAcid.getNuclio().size(); i++) {
        tryAddNuclio(originAcid.getNuclio()[i]);
    }
}

int AminoAcid::tryAddNuclio(const char nuclio) {
    if (addedNuclio.size() == 3) {
        return -3; // Если уже добавлено 3 нуклео
    } else if (nuclio == ' ') {
        return -2; // Если пустой символ
    }
    int incorrectNuclio = 0;
    for (int i = 0; i < NUCLIOTIDES.size(); i++) {
        if(nuclio != NUCLIOTIDES[i]) {
            incorrectNuclio++;
        }
    }
    if (incorrectNuclio == 4) {
        return -1; // Если несуществующий нуклео
    } else {
        addedNuclio.push_back(nuclio);
        return 0; // Если все ок
    }
}

void AminoAcid::addNuclio(const char nuclio) {
    int result = tryAddNuclio(nuclio);
    if (result == -3) {
        cout << "Вы добавили максимум нуклеотидов! Невозможно добавить еще!" << endl;
    } else if (result == -2) {
        cout << "Нельзя добавить пустоту!" << endl;
    } else if (result == -1) {
        cout << "Вы пытаетесь добавить несуществующий нуклеотид!" << endl;
    } else {
        cout << "Нуклеотид '" << nuclio << "' добавлен!"<< endl;
    }
}

bool AminoAcid::tryCombineInAcid() {
    if(addedNuclio.size() != 3) {
        return false;
    }
    totalCode = string() + addedNuclio[0] + addedNuclio[1] + addedNuclio[2]; 
    return true;
}

void AminoAcid::combineInAcid() {
    if (tryCombineInAcid()){
        cout << "Получилась аминокислота с кодом: " << getCode() << endl;
    } else {
        cout << "Не хватает нуклеотидов для объединения!" << endl;
    }
}

string AminoAcid::getCode() const {
    return totalCode;
}

void AminoAcid::chowCode() const {
    cout << "Кислота имеет код: " << totalCode << endl;
}

vector <char> AminoAcid::getNuclio() const {
    vector <char> result;
    for (int i = 0; i < addedNuclio.size(); i++) {
        result.push_back(addedNuclio[i]);
    }
    return result;
}

void AminoAcid::chowAddedNuclio() const {
    cout << "В кислоту уже добавлены: ";
    for (int i = 0; i < addedNuclio.size(); i++) {
        cout << addedNuclio[i];
    }
}

class ProteinSequence: public AminoAcid {
    public:
        ProteinSequence();
        ProteinSequence(const  vector <char> &newNuclio); //Конструктор инициализации 1
        ProteinSequence(const string &newCode); //Конструктор инициализации 2
        ProteinSequence(const ProteinSequence &originSeq); //Конструктор копирования
        void chowCode() const override;
        void addAcid(AminoAcid &acid);
        void addNuclios(const vector <char> &nuclios);
        string identification() const override;
    protected:
        int tryAddNuclio(const char nuclio) override;
    private:
        string totalCode = ""; //Фрагменты, добавленные в последовательность белка
};

string ProteinSequence::identification() const {
    return "ProteinSequence";
}

ProteinSequence::ProteinSequence() {}

ProteinSequence::ProteinSequence(const ProteinSequence &originSeq){
    totalCode = originSeq.getCode();
}

ProteinSequence::ProteinSequence(const vector <char> &newNuclio){
    int numberNuclio = 0;
    while(totalCode.length() != 20 && numberNuclio != newNuclio.size()){
        tryAddNuclio(newNuclio[numberNuclio]);
        numberNuclio++;
    }
    if (totalCode.length() == 0){
        cout << "Белковая последовательность создана, но из заданных нуклеотидов не добавлен ни один!" << endl;
    } else{
        cout << "Белковая последовательность создана. Она состоит из " << totalCode.length() << " элементов: ";
        cout << totalCode << endl;
    }
}

ProteinSequence::ProteinSequence(const string &newCode){
    int numberNuclio = 0;
    while(totalCode.length() != 20 && numberNuclio != newCode.length()){
        tryAddNuclio(newCode[numberNuclio]);
        numberNuclio++;
    }
    if (totalCode.length() == 0){
        cout << "Белковая последовательность создана, но из заданной последовательности не добавлен ни один нуклеотид!" << endl;
    } else{
        cout << "Белковая последовательность создана. Она состоит из " << totalCode.length() << " элементов: ";
        cout << totalCode << endl;
    }
}

int ProteinSequence::tryAddNuclio(const char nuclio) {
    if (totalCode.length() == 20) {
        return -3; // Если уже добавлено 20 частей
    } else if (nuclio == ' ') {
        return -2; // Если пустой символ
    }
    int incorrectNuclio = 0;
    for (int i = 0; i < NUCLIOTIDES.size(); i++) {
        if(nuclio != NUCLIOTIDES[i]) {
            incorrectNuclio++;
        }
    }
    if (incorrectNuclio == 4) {
        return -1; // Если несуществующий нуклео
    } else {
        totalCode = totalCode + nuclio;
        return 0; // Если все ок
    }   
}

void ProteinSequence::chowCode() const {
    cout << "Белковая последовательность имеет код: " << totalCode << endl;
}

void ProteinSequence::addAcid(AminoAcid &acid) {
    if (totalCode.length() >=18) {
        cout << "Невозможно добавить кислоту. Не хватает места!" << endl;
    } else {
        if (totalCode == ""){
            cout << "Невозможно добавить кислоту. Кислота не объединена!" << endl;
        } else {
            totalCode = totalCode + acid.getCode();
            cout << "Кислота добавлена в белковую последовательность." << endl;
        }
    }
}

void ProteinSequence::addNuclios(const vector <char> &nuclios) {
    if (totalCode.length() == 20) {
        cout << "Добавить нуклеотиды невозможно. Нет места." << endl;
    } else {
        int i = 0;
        int previousLenght = totalCode.length();
        while (totalCode.length() != 20 && i < nuclios.size()){
            tryAddNuclio(nuclios[i]);
            i++;
        }
        if (totalCode.length() != previousLenght){
            cout << "Добавлено " << totalCode.length() - previousLenght <<  " нуклеотидов в последовательность." << endl;
        } else {
            cout << "Ни один нуклеотид не добавлен." << endl;
        }
    }  
}

int main() { 
    cout << "Запуск тестов..." << endl;
    // Проверка конструктора
    AminoAcid acid1;
    assert(acid1.getCode().empty());
    assert(acid1.getNuclio().empty());
    cout << "Тест 1 - Done!" << endl;
    // Проверка копирования
    acid1.addNuclio('U');  
    AminoAcid acid2(acid1);
    assert(acid1.getCode() == acid2.getCode());
    assert(acid1.getNuclio().size() == acid2.getNuclio().size());
    for (int i = 0; i < acid2.getNuclio().size(); i++) {
        assert(acid1.getNuclio()[i] == acid2.getNuclio()[i]);
    }
    cout << "Тест 2 - Done!" << endl;
    // Проверка инициализации
    vector <char> nuclios = {'Q', ' ', 'A', 'U', 'G', 'A'};
    AminoAcid acid3 (nuclios);
    assert(acid3.getNuclio()[0] == 'A');
    assert(acid3.getNuclio()[1] == 'U');
    assert(acid3.getNuclio()[2] == 'G');
    cout << "Тест 3 - Done!" << endl;
    ProteinSequence seq1;
    assert(seq1.getCode().empty());
    cout << "Тест 4 - Done!" << endl;
    seq1.addNuclio('U');
    ProteinSequence seq2(seq1);
    assert(seq1.getCode() == seq2.getCode());
    cout << "Тест 5 - Done!" << endl;
    ProteinSequence seq3(nuclios);
    cout << "Тесты выполнены!" << endl << endl;
    
    AminoAcid newAcid;
    newAcid.addNuclio('U');
    newAcid.addNuclio('G');
    newAcid.addNuclio('Q');
    newAcid.addNuclio('G');
    newAcid.combineInAcid();

    ProteinSequence newSeq;
    newSeq.addNuclio('U');
    newSeq.addNuclio('A');
    newSeq.addNuclio('G');
    newSeq.chowCode();
    newSeq.addAcid(newAcid);
    newSeq.chowCode();
    
    return 0;
}